#include "CloudComm.h"
#include "TimingSingleton.h"
#include "SecureRandom.h"
#include "IoTString.h"
#include "Error.h"
#include "URL.h"
#include "Mac.h"

/**
 * Empty Constructor needed for child class.
 */
CloudComm::CloudComm() :
	baseurl(NULL),
	key(NULL),
	mac(NULL),
	password(NULL),
	random(NULL),
	salt(NULL),
	table(NULL),
	listeningPort(-1),
	localServerThread(NULL),
	doEnd(false),
	timer(TimingSingleton_getInstance())
{
}

/**
 * Constructor for actual use. Takes in the url and password.
 */
CloudComm::CloudComm(Table *_table,  IoTString *_baseurl, IoTString *_password, int _listeningPort) :
	baseurl(_baseurl),
	key(NULL),
	mac(NULL),
	password(_password),
	random(new SecureRandom()),
	salt(NULL),
	table(_table),
	listeningPort(_listeningPort),
	localServerThread(NULL),
	doEnd(false),
	timer(TimingSingleton_getInstance()) {
	if (listeningPort > 0) {
		localServerThread = new Thread(new Runnable() {
			void run() {
				localServerWorkerFunction();
			}
		});
		localServerThread->start();
	}
}

/**
 * Generates Key from password.
 */
SecretKeySpec *CloudComm::initKey() {
	try {
		PBEKeySpec *keyspec = new PBEKeySpec(password->internalBytes(),
																				 salt,
																				 65536,
																				 128);
		SecretKey *tmpkey = SecretKeyFactory_getInstance("PBKDF2WithHmacSHA256")->generateSecret(keyspec);
		return new SecretKeySpec(tmpkey->getEncoded(), "AES");
	} catch (Exception *e) {
		throw new Error("Failed generating key.");
	}
}

/**
 * Inits all the security stuff
 */

void CloudComm::initSecurity() {
	// try to get the salt and if one does not exist set one
	if (!getSalt()) {
		//Set the salt
		setSalt();
	}

	initCrypt();
}

/**
 * Inits the HMAC generator.
 */
void CloudComm::initCrypt() {
	if (password == NULL) {
		return;
	}

	try {
		key = initKey();
		password = NULL;// drop password
		mac = Mac_getInstance("HmacSHA256");
		mac->init(key);
	} catch (Exception *e) {
		throw new Error("Failed To Initialize Ciphers");
	}
}

/*
 * Builds the URL for the given request.
 */
URL *CloudComm::buildRequest(bool isput, int64_t sequencenumber, int64_t maxentries) {
	IoTString *reqstring = isput ? "req=putslot" : "req=getslot";
	IoTString *urlstr = baseurl + "?" + reqstring + "&seq=" + sequencenumber;
	if (maxentries != 0)
		urlstr += "&max=" + maxentries;
	return new URL(urlstr);
}

void CloudComm::setSalt() {

	if (salt != NULL) {
		// Salt already sent to server so dont set it again
		return;
	}

	try {
		Array<char> *saltTmp = new Array<char>(CloudComm_SALT_SIZE);
		random->nextBytes(saltTmp);

		for (int i = 0; i < CloudComm_SALT_SIZE; i++) {
			printf("%d\n", (int)saltTmp->get(i) & 255);
		}

		URL *url = new URL(baseurl + "?req=setsalt");
		timer->startTime();
		URLConnection *con = url->openConnection();
		HttpURLConnection *http = (HttpURLConnection *) con;

		http->setRequestMethod("POST");
		http->setFixedLengthStreamingMode(saltTmp->length());
		http->setDoOutput(true);
		http->setConnectTimeout(CloudComm_TIMEOUT_MILLIS);
		http->connect();

		OutputStream *os = http->getOutputStream();
		os->write(saltTmp);
		os->flush();

		int responsecode = http->getResponseCode();
		if (responsecode != HttpURLConnection_HTTP_OK) {
			// TODO: Remove this print
			printf("%d\n", responsecode);
			throw new Error("Invalid response");
		}

		timer->endTime();
		salt = saltTmp;
	} catch (Exception *e) {
		timer->endTime();
		throw new ServerException("Failed setting salt", ServerException_TypeConnectTimeout);
	}
}

bool CloudComm::getSalt() {
	URL *url = NULL;
	URLConnection *con = NULL;
	HttpURLConnection *http = NULL;

	try {
		url = new URL(baseurl + "?req=getsalt");
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}
	try {
		timer->startTime();
		con = url->openConnection();
		http = (HttpURLConnection *) con;
		http->setRequestMethod("POST");
		http->setConnectTimeout(CloudComm_TIMEOUT_MILLIS);
		http->setReadTimeout(CloudComm_TIMEOUT_MILLIS);
		http->connect();
		timer->endTime();
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSalt failed", ServerException_TypeConnectTimeout);
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}

	try {
		timer->startTime();
		int responsecode = http->getResponseCode();
		if (responsecode != HttpURLConnection_HTTP_OK) {
			throw new Error("Invalid response");
		}
		InputStream *is = http->getInputStream();
		if (is->available() > 0) {
			DataInputStream *dis = new DataInputStream(is);
			int salt_length = dis->readInt();
			Array<char> *tmp = new Array<char>(salt_length);
			dis->readFully(tmp);
			salt = tmp;
			timer->endTime();
			return true;
		} else {
			timer->endTime();
			return false;
		}
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSalt failed", ServerException_TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("getSlot failed");
	}
}

Array<char> *CloudComm::createIV(int64_t machineId, int64_t localSequenceNumber) {
	ByteBuffer *buffer = ByteBuffer_allocate(CloudComm_IV_SIZE);
	buffer->putLong(machineId);
	int64_t localSequenceNumberShifted = localSequenceNumber << 16;
	buffer->putLong(localSequenceNumberShifted);
	return buffer->array();
}

Array<char> *CloudComm::encryptSlotAndPrependIV(Array<char> *rawData, Array<char> *ivBytes) {
	try {
		IvParameterSpec *ivSpec = new IvParameterSpec(ivBytes);
		Cipher *cipher = Cipher_getInstance("AES/CTR/NoPadding");
		cipher->init(Cipher_ENCRYPT_MODE, key, ivSpec);

		Array<char> *encryptedBytes = cipher->doFinal(rawData);

		Array<char> *chars = new Array<char>(encryptedBytes->length() + CloudComm_IV_SIZE);
		System_arraycopy(ivBytes, 0, chars, 0, ivBytes->length());
		System_arraycopy(encryptedBytes, 0, chars, CloudComm_IV_SIZE, encryptedBytes->length());

		return chars;
	} catch (Exception *e) {
		throw new Error("Failed To Encrypt");
	}
}

Array<char> *CloudComm::stripIVAndDecryptSlot(Array<char> *rawData) {
	try {
		Array<char> *ivBytes = new Array<char>(CloudComm_IV_SIZE);
		Array<char> *encryptedBytes = new Array<char>(rawData->length() - CloudComm_IV_SIZE);
		System_arraycopy(rawData, 0, ivBytes, 0, CloudComm_IV_SIZE);
		System_arraycopy(rawData, CloudComm_IV_SIZE, encryptedBytes, 0, encryptedBytes->length);

		IvParameterSpec *ivSpec = new IvParameterSpec(ivBytes);
		Cipher *cipher = Cipher_getInstance("AES/CTR/NoPadding");
		cipher->init(Cipher_DECRYPT_MODE, key, ivSpec);
		return cipher->doFinal(encryptedBytes);
	} catch (Exception *e) {
		throw new Error("Failed To Decrypt");
	}
}

/*
 * API for putting a slot into the queue.  Returns NULL on success.
 * On failure, the server will send slots with newer sequence
 * numbers.
 */
Array<Slot *> *CloudComm::putSlot(Slot *slot, int max) {
	try {
		if (salt == NULL) {
			if (!getSalt()) {
				throw new ServerException("putSlot failed", ServerException_TypeSalt);
			}
			initCrypt();
		}

		int64_t sequencenumber = slot->getSequenceNumber();
		Array<char> *slotBytes = slot->encode(mac);
		Array<char> *chars = encryptSlotAndPrependIV(slotBytes, slot->getSlotCryptIV());
		URL *url = buildRequest(true, sequencenumber, max);
		timer->startTime();
		URLConnection *con = url->openConnection();
		HttpURLConnection *http = (HttpURLConnection *) con;
		http->setRequestMethod("POST");
		http->setFixedLengthStreamingMode(chars->length);
		http->setDoOutput(true);
		http->setConnectTimeout(CloudComm_TIMEOUT_MILLIS);
		http->setReadTimeout(CloudComm_TIMEOUT_MILLIS);
		http->connect();
		OutputStream *os = http->getOutputStream();
		os->write(chars);
		os->flush();
		timer->endTime();
	} catch (ServerException *e) {
		timer->endTime();
		throw e;
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("putSlot failed", ServerException_TypeConnectTimeout);
	} catch (Exception *e) {
		throw new Error("putSlot failed");
	}

	try {
		timer->startTime();
		InputStream *is = http->getInputStream();
		DataInputStream *dis = new DataInputStream(is);
		Array<char> *resptype = new Array<char>(7);
		dis->readFully(resptype);
		timer->endTime();

		if (Arrays->equals(resptype, "getslot"->getBytes())) {
			return processSlots(dis);
		} else if (Arrays->equals(resptype, "putslot"->getBytes())) {
			return NULL;
		} else
			throw new Error("Bad response to putslot");
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("putSlot failed", ServerException->TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("putSlot failed");
	}
}

/**
 * Request the server to send all slots with the given
 * sequencenumber or newer->
 */
Array<Slot *> *CloudComm::getSlots(int64_t sequencenumber) {
	try {
		if (salt == NULL) {
			if (!getSalt()) {
				throw new ServerException("getSlots failed", ServerException_TypeSalt);
			}
			initCrypt();
		}

		URL *url = buildRequest(false, sequencenumber, 0);
		timer->startTime();
		URLConnection *con = url->openConnection();
		HttpURLConnection *http = (HttpURLConnection *) con;
		http->setRequestMethod("POST");
		http->setConnectTimeout(CloudComm_TIMEOUT_MILLIS);
		http->setReadTimeout(CloudComm_TIMEOUT_MILLIS);
		http->connect();
		timer->endTime();
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSlots failed", ServerException_TypeConnectTimeout);
	} catch (ServerException *e) {
		timer->endTime();

		throw e;
	} catch (Exception *e) {
		throw new Error("getSlots failed");
	}

	try {
		timer->startTime();
		InputStream *is = http->getInputStream();
		DataInputStream *dis = new DataInputStream(is);
		Array<char> *resptype = new Array<char>(7);
		dis->readFully(resptype);
		timer->endTime();
		if (!resptype->equals("getslot"->getBytes()))
			throw new Error("Bad Response: " + new String(resptype));

		return processSlots(dis);
	} catch (SocketTimeoutException *e) {
		timer->endTime();
		throw new ServerException("getSlots failed", ServerException_TypeInputTimeout);
	} catch (Exception *e) {
		throw new Error("getSlots failed");
	}
}

/**
 * Method that actually handles building Slot objects from the
 * server response.  Shared by both putSlot and getSlots.
 */
Array<Slot *> *CloudComm::processSlots(DataInputStream *dis) {
	int numberofslots = dis->readInt();
	Array<int> *sizesofslots = new Array<int>(numberofslots);
	Array<Slot *> *slots = new Array<Slot *>(numberofslots);

	for (int i = 0; i < numberofslots; i++)
		sizesofslots->set(i, dis->readInt());
	for (int i = 0; i < numberofslots; i++) {
		Array<char> *rawData = new Array<char>(sizesofslots->get(i));
		dis->readFully(rawData);
		Array<char> *data = stripIVAndDecryptSlot(rawData);
		slots->set(i, Slot_decode(table, data, mac));
	}
	dis->close();
	return slots;
}

Array<char> *sendLocalData(Array<char> *sendData, int64_t localSequenceNumber, String host, int port) {
	if (salt == NULL)
		return NULL;
	try {
		printf("Passing Locally\n");
		mac->update(sendData);
		Array<char> *genmac = mac->doFinal();
		Array<char> *totalData = new Array<char>(sendData->length() + genmac->length());
		System_arraycopy(sendData, 0, totalData, 0, sendData->length());
		System_arraycopy(genmac, 0, totalData, sendData->length(), genmac->length());

		// Encrypt the data for sending
		Array<char> *iv = createIV(table->getMachineId(), table->getLocalSequenceNumber());
		Array<char> *encryptedData = encryptSlotAndPrependIV(totalData, iv);

		// Open a TCP socket connection to a local device
		Socket *socket = new Socket(host, port);
		socket->setReuseAddress(true);
		DataOutputStream *output = new DataOutputStream(socket->getOutputStream());
		DataInputStream *input = new DataInputStream(socket->getInputStream());

		timer->startTime();
		// Send data to output (length of data, the data)
		output->writeInt(encryptedData->length);
		output->write(encryptedData, 0, encryptedData->length);
		output->flush();

		int lengthOfReturnData = input->readInt();
		Array<char> *returnData = new Array<char>(lengthOfReturnData);
		input->readFully(returnData);
		timer->endTime();
		returnData = stripIVAndDecryptSlot(returnData);

		// We are done with this socket
		socket->close();
		mac->update(returnData, 0, returnData->length - HMAC_SIZE);
		Array<char> *realmac = mac->doFinal();
		Array<char> *recmac = new Array<char>(HMAC_SIZE);
		System_arraycopy(returnData, returnData->length - realmac->length, recmac, 0, realmac->length);

		if (!recmac->equals(realmac))
			throw new Error("Local Error: Invalid HMAC!  Potential Attack!");

		Array<char> *returnData2 = new Array<char>(lengthOfReturnData - recmac->length());
		System_arraycopy(returnData, 0, returnData2, 0, returnData2->length);

		return returnData2;
	} catch (Exception *e) {
		printf("Exception\n");
	}

	return NULL;
}

void CloudComm::localServerWorkerFunction() {
	ServerSocket *inputSocket = NULL;

	try {
		// Local server socket
		inputSocket = new ServerSocket(listeningPort);
		inputSocket->setReuseAddress(true);
		inputSocket->setSoTimeout(CloudComm_TIMEOUT_MILLIS);
	} catch (Exception *e) {
		throw new Error("Local server setup failure...");
	}

	while (!doEnd) {
		try {
			// Accept incoming socket
			Socket *socket = inputSocket->accept();
			DataInputStream *input = new DataInputStream(socket->getInputStream());
			DataOutputStream *output = new DataOutputStream(socket->getOutputStream());

			// Get the encrypted data from the server
			int dataSize = input->readInt();
			Array<char> *readData = new Array<char>(dataSize);
			input->readFully(readData);
			timer->endTime();

			// Decrypt the data
			readData = stripIVAndDecryptSlot(readData);
			mac->update(readData, 0, readData->length - HMAC_SIZE);
			Array<char> *genmac = mac->doFinal();
			Array<char> *recmac = new Array<char>(HMAC_SIZE);
			System_arraycopy(readData, readData->length() - recmac->length(), recmac, 0, recmac->length());

			if (!recmac->equals(genmac))
				throw new Error("Local Error: Invalid HMAC!  Potential Attack!");

			Array<char> *returnData = new Array<char>(readData->length() - recmac->length());
			System_arraycopy(readData, 0, returnData, 0, returnData->length());

			// Process the data
			Array<char> *sendData = table->acceptDataFromLocal(returnData);
			mac->update(sendData);
			Array<char> *realmac = mac->doFinal();
			Array<char> *totalData = new Array<char>(sendData->length() + realmac->length());
			System_arraycopy(sendData, 0, totalData, 0, sendData->length());
			System_arraycopy(realmac, 0, totalData, sendData->length(), realmac->length());

			// Encrypt the data for sending
			Array<char> *iv = createIV(table->getMachineId(), table->getLocalSequenceNumber());
			Array<char> *encryptedData = encryptSlotAndPrependIV(totalData, iv);

			timer->startTime();
			// Send data to output (length of data, the data)
			output->writeInt(encryptedData->length());
			output->write(encryptedData, 0, encryptedData->length());
			output->flush();

			// close the socket
			socket->close();
		} catch (Exception *e) {
		}
	}

	if (inputSocket != NULL) {
		try {
			inputSocket->close();
		} catch (Exception *e) {
			throw new Error("Local server close failure...");
		}
	}
}

void CloudComm::close() {
	doEnd = true;

	if (localServerThread != NULL) {
		try {
			localServerThread->join();
		} catch (Exception *e) {
			throw new Error("Local Server thread join issue...");
		}
	}
}
